/* ********************************************************
    color.cpp

    Author = Hideyuki Harada
    Nihon Univ. Tani Research Group


    Tomita's Algorithm for resolving MaxClique problem
    designed for "g++ 3.3.3" with "LEDA 4.5"
   ******************************************************** */


#include<LEDA/graph.h>
#include<LEDA/list.h>
#include<time.h>
#include<iostream.h>
#include<fstream.h>
#include<string.h>

#define P_VERSION "color 0.01"

using namespace leda;

graph G;           // undirected graph

list<node> Qmax;   // node set for MaxClique

int Branch_count;  // number of counted SST branch

int coloring_count;

int N[10000];

void read_graph(int argc, char *argv[]);   // read filename from command line

void initialize(list<node> *R, list<node> *Q);  // clear Q and put V to R

void show_result(float T, char *argv[]);  // write result

void dfmax(list<node> R, list<node> Q);  // basic algorithm

int coloring(list<node> Rp);

void sort_by_degree(list<node> R);

int cmp_deg(const node &v, const node &u);

//*************************************************


int main(int argc, char *argv[]){

  float T = used_time();
  list<node> R, Q;

  if(argc != 2){
    cout << "usage: dfmax [file]" << endl;
    exit(1);
  }

  read_graph(argc, argv);

  initialize(&R, &Q);


  // start procedure
  dfmax(R, Q);


  show_result(T, argv);

  return 0;

}


void dfmax(list<node> R, list<node> Q){

  int c;
  node p, v;
  list<node> Rp;

  while(!R.empty()){

    Branch_count++;

    p = R.head();

    Q.append(p);

    Rp.clear();
    // Rp = intersect(R, G.adj_nodes(p));
    forall_adj_nodes(v, p){
      if(R.search(v) != NULL){
	Rp.append(v);
      }
    }

    if(!Rp.empty()){
      if( Q.size()+coloring(Rp) > Qmax.size()){
        dfmax(Rp, Q);
      }
    }
    else{
      if(Q.size() > Qmax.size()){  // apply new MaxClique
	Qmax.clear();
	forall(v, Q){
	  Qmax.append(v);
	}
      }
    }
    
    Q.remove(p);
    R.remove(p);
  }

  return;

}


int coloring(list<node> Rp){

  int c, k, flag;
  list<node> C[500];
  node p, v;

  c=0;
  while(!Rp.empty()){

    p = Rp.head();

    k=0;
    
    flag = 0;
    do{
      forall(v, C[k]){
	coloring_count ++;
	if(((G.adj_nodes(p)).search(v)) != NULL){
	  flag = 1;
	}
      }
      k++;
    }while(flag == 1);
    
    if(k > c){
      c = k;
    }
    C[k].append(p);

    
    Rp.remove(p);
  }
  // color(v) = min{}


  return c;
}


void sort_by_degree(list<node> R){
  

  R.sort(*cmp_deg, 1);


  return;

}

int cmp_deg(const node &v, const node &u){

  if(degree(v) < degree(u)){
    return -1;
  }

  else
    if(degree(v) == degree(u)){
      return 0;
    }

    else
      return 1;
}

void read_graph(int argc, char *argv[]){

  G.read(argv[1]);
  G.make_undirected();

  return;
}

void initialize(list<node> *R, list<node> *Q){

  (*Q).clear();
  (*R) = G.all_nodes();
  Qmax.clear();
  Branch_count = 0;
  coloring_count =0;

  return;
}

void show_result(float T, char *argv[]){

  node v;
  ofstream fout;
  time_t t=time(NULL);
  string ofile;

  ofile = ".result";
 
  fout.open(argv[1]+ofile, ios::app);

  fout << asctime(localtime(&t));
  fout << "Program version = " << P_VERSION << endl;  
  fout << "Filename = " << argv[1] << endl;
  fout << " Number of vertexes = " << G.number_of_nodes() << endl;
  fout << " Number of edges = " << G.number_of_edges() << endl;
  fout << "Used time = " << used_time(T) << endl;
  fout << "SST Branch count = " << Branch_count << endl;
  fout << "coloring node count = " << coloring_count << endl;
  fout << "MAX CLIQUE SIZE = " << Qmax.size() << endl;
  fout << "CLIQUE ELEMENT(S) :" << endl;

  forall(v, Qmax){
    fout << " " << index(v);
  }
  fout << endl;

  fout << endl;

  fout.close();
 
  return;
}
